"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Repository = void 0;
const eventemitter3_1 = require("eventemitter3");
const persistenceLayer_1 = require("./persistenceLayer");
const Event_1 = require("./Event");
const typeorm_1 = require("typeorm");
const debug_1 = __importDefault(require("debug"));
const log = (0, debug_1.default)("sourced-repo-typeorm");
class Repository extends eventemitter3_1.EventEmitter {
    constructor(EntityType, options = {
        indices: [],
        forceSnapshot: false,
        snapshotFrequency: 10,
    }) {
        super();
        if (!persistenceLayer_1.persistenceLayer.connection) {
            throw new Error("🚨 persistenceLayer has not been initialized. you must call require('sourced-repo/persistenceLayer').connect(connectionOptions); before instantiating a Repository");
        }
        const connection = persistenceLayer_1.persistenceLayer.connection;
        this.connection = connection;
        const indices = [...new Set([...options.indices, ...["id"]])];
        this.EntityType = EntityType;
        this.indices = indices;
        this.snapshotFrequency = options.snapshotFrequency;
        this.forceSnapshot = options.forceSnapshot;
        this.events = this.connection.getRepository(Event_1.Event);
        log(`✅ "${this.EntityType.name}" events store ready`);
        this.emit("ready");
    }
    commit(entity, options) {
        return __awaiter(this, void 0, void 0, function* () {
            log(`⏳ committing ${this.EntityType.name} for id ${entity.id}`);
            yield this._commitEvents(entity, options);
            this._emitEvents(entity);
            return entity;
        });
    }
    get(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const entity = yield this._getByIndex("id", id);
            return entity;
        });
    }
    _getByIndex(index, value) {
        return __awaiter(this, void 0, void 0, function* () {
            log(`⏳ getting ${this.EntityType.name} where "${index}" is "${value}"`);
            const snapshot = yield this.events.findOne({
                order: {
                    version: "DESC",
                },
                where: [
                    {
                        snapshot: true,
                        entityType: this.EntityType.name,
                        [index]: value,
                    },
                ],
            });
            const events = yield this.events.find({
                order: {
                    version: "ASC",
                },
                where: [
                    {
                        snapshot: false,
                        entityType: this.EntityType.name,
                        [index]: value,
                        version: (0, typeorm_1.MoreThan)((snapshot === null || snapshot === void 0 ? void 0 : snapshot.version) || 0),
                    },
                ],
            });
            log(`⏳ results for ${this.EntityType.name} where "${index}" is "${value}"`, { snapshot: !!snapshot, events: events.length });
            if (!snapshot && !events.length) {
                return null;
            }
            const id = index === "id" ? value : snapshot ? snapshot.id : events[0].id;
            const entity = this._deserialize(id, snapshot === null || snapshot === void 0 ? void 0 : snapshot.data, events);
            return entity;
        });
    }
    _commitEvents(entity, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (entity.newEvents.length === 0)
                return null;
            if (!entity.id) {
                throw new Error(`🚨 Cannot commit an entity of type ${this.EntityType} without an [id] property`);
            }
            const newEvents = entity.newEvents;
            newEvents.forEach((event) => {
                this.indices.forEach(function (index) {
                    event[index] = entity[index];
                });
            });
            const eventObjects = newEvents.map((newEvent) => {
                const event = new Event_1.Event();
                event.id = entity.id;
                event.version = newEvent.version;
                event.snapshotVersion = entity.snapshotVersion;
                event.timestamp = newEvent.timestamp;
                event.method = newEvent.method;
                event.entityType = this.EntityType.name;
                event.data = newEvent.data;
                event.snapshot = false;
                return event;
            });
            if (this.forceSnapshot ||
                (options === null || options === void 0 ? void 0 : options.forceSnapshot) ||
                entity.version >= entity.snapshotVersion + this.snapshotFrequency) {
                const snapshot = entity.snapshot();
                const snapshotEvent = new Event_1.Event();
                snapshotEvent.snapshot = true;
                snapshotEvent.method = "snapshot";
                snapshotEvent.data = snapshot;
                snapshotEvent.id = snapshot.id;
                snapshotEvent.version = snapshot.version;
                snapshotEvent.snapshotVersion = snapshot.snapshotVersion;
                snapshotEvent.timestamp = snapshot.timestamp;
                snapshotEvent.entityType = this.EntityType.name;
                eventObjects.push(snapshotEvent);
            }
            log("⏳ Inserting event objects", { eventObjects: eventObjects.length });
            yield this.events.insert(eventObjects);
            entity.newEvents = [];
            log(`✅ committed ${this.EntityType.name}.events for id ${entity.id}`);
            return entity;
        });
    }
    _deserialize(id, snapshot, events) {
        log("⏳ deserializing %s entity ", this.EntityType.name);
        const entity = new this.EntityType(snapshot, events);
        entity.id = id;
        return entity;
    }
    _emitEvents(entity) {
        const eventsToEmit = entity.eventsToEmit;
        log("⏳ emitting events", { eventsToEmit: eventsToEmit.length });
        entity.eventsToEmit = [];
        eventsToEmit.forEach((eventToEmit) => {
            const args = Array.prototype.slice.call(eventToEmit);
            this.EntityType.prototype.emit.apply(entity, args);
        });
        log(`✅ emitted local events for id ${entity.id}`);
    }
}
exports.Repository = Repository;
